#include "common.h"
#include "gpl.h"

// #define DEBUG
//
#ifdef DEBUG
#define DEBUG_print(a) printf a;
#else
#define DEBUG_print(a)
#endif

FILE *DB;			// DB FH
struct stat DB_stat;		// DB file stat info
void *DB_map;			// DB mmap'd address
char *tempDB_map;		// Place to hold the DB temporarily
int uid;			// Calling UID
int euid;			// Effective UID
char *username;			// username

// can be set by processes that can be called as root
int root_caller_required = 0;

// Config file
const char *devzone_db = "/etc/devzone/user_zones.txt";

/* Config file format:
 * 80 chars zonename padded with space:12 chars padded with space:5 chars padded with space\n
 * zone-a                                                                          :username    :uid     
 *
 * zones that are not reserved have a blank username and the uid field is blank or zero
 */

// Returns !0 on success, 0 on failure
// Also prints errors to stderr (bad style but quickly accomplished)
// TODO: check "username" against uid
int
sanity_checks (int uid, int euid, char *username)
{
  struct passwd *passwd_entry;

  // Check calling and effective UIDs, make sure we are effectively root
  if (euid != 0)
    {
      fprintf (stderr,
	       "Error: not running with root permissions (euid==%d).\n",
	       euid);
      return 0;
    }

  if (username == NULL)
    {
      if (euid == uid && uid == 0)
	{
	  username = "root";
	}
      else
	{
	  fprintf (stderr, "Error: Bad Env: USER is NULL\n");
	  return -1;
	}
    }

  passwd_entry = getpwuid (uid);
  DEBUG_print (("uid: %d\n", uid));
  DEBUG_print (("passwd_entry: %d\n", passwd_entry));
  DEBUG_print (("pw_name: %s\n", passwd_entry->pw_name));
  DEBUG_print (("username(d): %d\n", username));
  DEBUG_print (("username(s): %s\n", username));
  if (strcmp (passwd_entry->pw_name, username))
    {
      fprintf (stderr, "Danger Will Robinson\n");
      return 0;
    }

// WARNING: enable debugging changes functionality of the program HERE
#ifdef DEBUG
  DEBUG_print (("Calling UID/Effective UID: %d/%d.\n", uid, euid));
#else // ! DEBUG
  // We don't want to actually BE root (unless we need to debug for some reason)
  if (uid == 0 && root_caller_required == 0)
    {
      fprintf (stderr, "Error: Do not run directly as root (uid==%d).\n",
	       euid);
      return 0;
    }
  else if (uid != 0 && root_caller_required != 0)
    {
      fprintf (stderr, "Error: Needs to be run as root\n");
      return 0;
    }
#endif // ! DEBUG

  return 1;
}

void
init_DB ()
{

  // Open and lock db file
  DB = fopen (devzone_db, "r+");
  if (!DB)
    {
      // try to create db if it doesn't exist
      DB = fopen (devzone_db, "a");
      if (DB)
	{
	  fclose (DB);
	}
      DB = fopen (devzone_db, "r+");
      if (!DB)
	{
	  perror ("Could not open DB file");
	  exit (-1);
	}
    }
  lockf (fileno (DB), F_LOCK, 0);
  // MMAP file for easy editing
  DB_map =
    mmap (NULL, 1048576, PROT_READ | PROT_WRITE, MAP_SHARED, fileno (DB), 0);
  if (!DB_map)
    {
      perror ("Could not mmap DB file");
      exit (-1);
    }
  fstat (fileno (DB), &DB_stat);

}

void
close_DB ()
{

  // munmap
  lockf (fileno (DB), F_ULOCK, 0);
  fclose (DB);

}

// exits on failure
void
init ()
{
  DEBUG_print (("+init()\n"));

  // Ask system for calling/effective UIDs
  uid = getuid ();
  euid = geteuid ();
  // Ask the environment for a username (checked in sanity_checks())
  username = getenv ("USER");

  if (!sanity_checks (uid, euid, username))
    {
      exit (-1);
    }

  init_DB ();

  DEBUG_print (("-init()\n"));
}

// exits on failure
void
destroy ()
{
  close_DB ();
}

int
getZoneRecordCount ()
{
  return DB_stat.st_size / CONFIG_LINE_LEN;
}

// setup the zone info reference passed in based on
// the position of the record
void
getZoneRecordByID (const int record_id, struct zone_info *ret)
{
  int i;			// iterate in a loop

  memcpy (ret, DB_map + CONFIG_LINE_LEN * record_id, CONFIG_LINE_LEN);

  // NULL out spaces to parse strings easily
  for (i = 0; i < CONFIG_ZONE_LEN; i++)
    if (ret->name[i] == ' ')
      ret->name[i] = '\0';
  for (i = 0; i < CONFIG_USER_LEN; i++)
    if (ret->username[i] == ' ')
      ret->username[i] = '\0';
  for (i = 0; i < CONFIG_UID_LEN; i++)
    if (ret->char_uid[i] == ' ')
      ret->char_uid[i] = '\0';

  // if we have a valid uid, parse it into a number
  if (ret->char_uid[0] != '\0')
    {
      ret->uid = atoi (ret->char_uid);
    }
  else
    {
      ret->uid = -1;
    }
}

int
getZoneRecordByName (char *zone_name, struct zone_info *ret)
{
  int record_id, i;
  DEBUG_print (("+getZoneRecordByName(%s, ...)\n", zone_name));

  for (record_id = getZoneRecordCount (); record_id >= 0; record_id--)
    {
      memcpy (ret, DB_map + CONFIG_LINE_LEN * record_id, CONFIG_LINE_LEN);
      // NULL out spaces to parse strings easily
      for (i = 0; i < CONFIG_ZONE_LEN; i++)
	if (ret->name[i] == ' ')
	  ret->name[i] = '\0';
      for (i = 0; i < CONFIG_USER_LEN; i++)
	if (ret->username[i] == ' ')
	  ret->username[i] = '\0';
      for (i = 0; i < CONFIG_UID_LEN; i++)
	if (ret->char_uid[i] == ' ')
	  ret->char_uid[i] = '\0';

      // if we have a valid uid, parse it into a number
      if (ret->char_uid[0] != '\0')
	{
	  ret->uid = atoi (ret->char_uid);
	}
      else
	{
	  ret->uid = -1;
	}

      // whole name matches? .. let's return!
      if (strcmp (zone_name, ret->name) == 0)
	{
	  DEBUG_print (("-getZoneRecordByName(%s): 0 = found record\n",
			zone_name));
	  return 0;
	}
    }

  // Couldn't find a zone with the given name
  ret->name[0] = '\0';
  ret->username[0] = '\0';
  ret->char_uid[0] = '\0';
  DEBUG_print (("-getZoneRecordByName(%s): -1 = record not found\n",
		zone_name));
  return -1;
}

// returns a newly malloc'd list (NULL terminated) of
// zone_info structs corresponding to the DB list
// caller is responsible for free'ing memory
struct zone_info **
getRecordList ()
{
  int count, i;
  struct zone_info **ret;
  struct zone_info tmp;

  count = getZoneRecordCount ();

  // Allocate a bunch of objects
  ret =
    (struct zone_info **) malloc (sizeof (struct zone_info *) * (count + 1));
  for (i = count - 1; i >= 0; i--)
    {
      ret[i] = (struct zone_info *) malloc (sizeof (struct zone_info));
    }

  DEBUG_print (("getRecordList(): Allocated %d chunks of %d size\n", (count),
		sizeof (struct zone_info)));

  // Null terminated list
  ret[count] = NULL;

  for (count--; count >= 0; count--)
    {
      getZoneRecordByID (count, &tmp);
      DEBUG_print (("Copy zone id %d\n", count));
      memcpy (ret[count], &tmp, sizeof (struct zone_info));
    }

  return ret;
}

// Let's be paranoid...
void
updateRecordID (int recordID, struct zone_info *newInfo)
{
  char newLine[CONFIG_LINE_LEN + 1];	// to be copied into the config
  int i;

  DEBUG_print (("+updateRecordID(%d, ...)\n", recordID));

  for (i = 0; i <= CONFIG_LINE_LEN; i++)
    newLine[i] = ' ';

  memcpy (newLine, newInfo->name, strlen (newInfo->name));
  memcpy (newLine + CONFIG_ZONE_LEN + 1, newInfo->username,
	  strlen (newInfo->username));
  memcpy (newLine + CONFIG_ZONE_LEN + CONFIG_USER_LEN + 2, newInfo->char_uid,
	  strlen (newInfo->char_uid));

  for (i = 0; i <= CONFIG_LINE_LEN; i++)
    if (newLine[i] == '\0')
      newLine[i] = ' ';

  newLine[CONFIG_ZONE_LEN] = CONFIG_SEP;
  newLine[CONFIG_ZONE_LEN + 1 + CONFIG_USER_LEN] = CONFIG_SEP;
  newLine[CONFIG_LINE_LEN - 1] = '\n';
  newLine[CONFIG_LINE_LEN + 1] = '\0';
  DEBUG_print (("new DB line: %s", newLine));
  memcpy (DB_map + CONFIG_LINE_LEN * recordID, newLine, CONFIG_LINE_LEN);
  msync (DB_map + CONFIG_LINE_LEN * recordID, CONFIG_LINE_LEN, MS_ASYNC);

  DEBUG_print (("-updateRecordID(%d, ...)\n", recordID));
}

// semi-paranoid again, in-efficient
int
addRecord (struct zone_info *newInfo)
{
  int i;
  char newLine[CONFIG_LINE_LEN + 1];
  DEBUG_print (("+addRecord()"));

  for (i = 0; i <= CONFIG_LINE_LEN; i++)
    newLine[i] = ' ';

  memcpy (newLine, newInfo->name, strlen (newInfo->name));
  memcpy (newLine + CONFIG_ZONE_LEN + 1, newInfo->username,
	  strlen (newInfo->username));
  memcpy (newLine + CONFIG_ZONE_LEN + CONFIG_USER_LEN + 2, newInfo->char_uid,
	  strlen (newInfo->char_uid));

  for (i = 0; i <= CONFIG_LINE_LEN; i++)
    if (newLine[i] == '\0')
      newLine[i] = ' ';

  newLine[CONFIG_ZONE_LEN] = CONFIG_SEP;
  newLine[CONFIG_ZONE_LEN + 1 + CONFIG_USER_LEN] = CONFIG_SEP;
  newLine[CONFIG_LINE_LEN - 1] = '\n';
  newLine[CONFIG_LINE_LEN] = '\0';
  DEBUG_print (("new DB line: %s", newLine));

  fseek (DB, 0, SEEK_END);
  fprintf (DB, "%s", newLine);

  // Yuck, but works
  close_DB ();
  init_DB ();

  DEBUG_print (("-addRecord()"));
  return 0;
}

int
delRecord (char *zoneName)
{
  int i, rec_count, range_size;
  void *rec_start;
  char notFound = 1;
  struct zone_info zone;

  DEBUG_print (("+delRecord(%s, ...)\n", zoneName));

  // Find the record id
  rec_count = getZoneRecordCount ();
  i = 0;
  while (++i <= rec_count && notFound)
    {
      getZoneRecordByID (i, &zone);
      DEBUG_print (("%s == %s\n", zone.name, zoneName));
      if (strcmp (zone.name, zoneName) == 0)
	{
	  DEBUG_print (("found record id=%d\n", i));
	  notFound = 0;
	}
    }

  // Did we find the record?
  if (notFound)
    {
      fprintf (stderr, "Error: zone not found in DB.\n");
      return -1;
    }

  // now i contains the record id to be removed
  // Let's figure out how to move some memory
  range_size = (rec_count - i) * CONFIG_LINE_LEN;

  // Not the last record? Let's move everything up 1 slot
  if (range_size > 0)
    {
      DEBUG_print (("move %d bytes (%d records) up 1 slot\n", range_size,
		    rec_count - i + 1));
      rec_start = DB_map + CONFIG_LINE_LEN * (i - 1);
      memmove (rec_start, DB_map + CONFIG_LINE_LEN * (i), range_size);
      msync (rec_start, range_size, MS_INVALIDATE);
    }

  DEBUG_print (("truncate file to %d bytes\n",
		CONFIG_LINE_LEN * (rec_count - 1)));
  ftruncate (fileno (DB), CONFIG_LINE_LEN * (rec_count - 1));

  // reset DB (mmap/etc)
  DEBUG_print (("reset DB\n"));
  close_DB ();
  init_DB ();

  DEBUG_print (("-delRecord(%s, ...)\n", zoneName));
  return 0;
}

char *
getReservedZoneName ()
{
  char *ret_zonename = NULL;
  int i = 0;
  struct zone_info currentZone;

  DEBUG_print (("+getReservedZoneName()\n"));

  for (i = 0; i < getZoneRecordCount (); i++)
    {
      getZoneRecordByID (i, &currentZone);

      DEBUG_print (("%s - %s (%d)\n", currentZone.name, currentZone.username,
		    currentZone.uid));

      // Did we find the reserved zone name?
      if (currentZone.uid == uid)
	{
	  ret_zonename =
	    (char *) malloc (sizeof (char) * (1 + strlen (currentZone.name)));
	  strcpy (ret_zonename, currentZone.name);

	  DEBUG_print (("-getReservedZoneName(): %s\n", ret_zonename));

	  return ret_zonename;
	}

    }

  DEBUG_print (("-getReservedZoneName(): NULL\n"));

  // We did not find the reserved zone
  return NULL;
}

char *
getFreeZoneName ()
{
  char *ret_zonename = NULL;
  int i = 0;
  struct zone_info currentZone;

  DEBUG_print (("+getFreeZoneName()\n"));

  for (i = 0; i < getZoneRecordCount (); i++)
    {
      getZoneRecordByID (i, &currentZone);

      DEBUG_print (("%s - %s (%d)\n", currentZone.name, currentZone.username,
		    currentZone.uid));

      // Did we find the reserved zone name?
      if (currentZone.uid == -1)
	{
	  ret_zonename =
	    (char *) malloc (sizeof (char) * (1 + strlen (currentZone.name)));
	  strcpy (ret_zonename, currentZone.name);

	  DEBUG_print (("-getFreeZoneName(): %s\n", ret_zonename));

	  return ret_zonename;
	}
    }

  DEBUG_print (("-getFreeZoneName(): NULL\n"));

  return NULL;
}

// helper for things that want this.
char *
getUsername ()
{
  return username;
}

char *
reserveZone (char *target_zone_name)
{
  char *ret_zonename = NULL;
  int i = 0;
  struct zone_info currentZone;

  DEBUG_print (("+reserveZone()\n"));

  for (i = 0; i < getZoneRecordCount (); i++)
    {
      getZoneRecordByID (i, &currentZone);

      DEBUG_print (("%s - %s (%d)\n", currentZone.name, currentZone.username,
		    currentZone.uid));

      if (strcmp (target_zone_name, currentZone.name) != 0)
	{
	  DEBUG_print (("no match...\n"));
	  continue;
	}

      DEBUG_print (("Found db record (%d), update line in file...\n", i));
      DEBUG_print (("%s:%s:%d\n", target_zone_name, username, uid));

      // Setup zone with new info
      memcpy (currentZone.username, username, strlen (username));
      currentZone.uid = uid;
      sprintf (currentZone.char_uid, "%d", uid);

      updateRecordID (i, &currentZone);

      ret_zonename =
	(char *) malloc (sizeof (char) * (1 + strlen (currentZone.name)));
      strcpy (ret_zonename, currentZone.name);
      DEBUG_print (("-reserveZone(): %s\n", ret_zonename));
      return ret_zonename;
    }

  DEBUG_print (("-reserveZone(): NULL\n"));
  return NULL;
}

char *
freeZone (char *target_zone_name)
{
  char *ret_zonename = NULL;
  int i = 0;
  struct zone_info currentZone;

  DEBUG_print (("+freeZone()\n"));

  for (i = 0; i < getZoneRecordCount (); i++)
    {
      getZoneRecordByID (i, &currentZone);

      DEBUG_print (("%s - %s (%d)\n", currentZone.name, currentZone.username,
		    currentZone.uid));

      if (strcmp (target_zone_name, currentZone.name) != 0)
	{
	  DEBUG_print (("no match...\n"));
	  continue;
	}

      DEBUG_print (("Found db record (%d), update line in file...\n", i));
      DEBUG_print (("%s:%s:%d\n", target_zone_name, username, uid));

      // Setup zone with new info
      currentZone.username[0] = '\0';
      currentZone.uid = -1;
      currentZone.char_uid[0] = '\0';

      updateRecordID (i, &currentZone);

      ret_zonename =
	(char *) malloc (sizeof (char) * (1 + strlen (currentZone.name)));
      strcpy (ret_zonename, currentZone.name);
      DEBUG_print (("-freeZone(): %s\n", ret_zonename));
      return ret_zonename;
    }

  DEBUG_print (("-freeZone(): NULL\n"));
  return NULL;
}

void
rootRequired ()
{
  DEBUG_print (("rootRequired()\n"));
  root_caller_required = 1;
}

int
getZoneState (const char *zoneName, char *zoneState)
{
  int childWritePipe[2];	// Allow children to talk back to parent
  char readIn[1024];		// read in up to 1k char line
  int readBytes = 0;
  int childPID;
  int exitStatus;
  int i;			// basic iteration

  zoneState[0] = '\0';

  // Allocate pipe for child->parent communication
  pipe (childWritePipe);

  childPID = fork ();
  if (!childPID)
    {				// child
      setreuid (0, 0);
      // Redirect stdout to parent pipe
      close (fileno (stdout));
      dup2 (childWritePipe[WRITE_END], fileno (stderr));
      dup2 (childWritePipe[WRITE_END], fileno (stdout));

      // Get zone status
      execl ("/usr/sbin/zoneadm", "zoneadm", "-z", zoneName, "list", "-p",
	     (char *) NULL);
      printf ("WE ARE HERE!!!\n");
    }
  wait (&exitStatus);

  // Get a line of data from the child process (kludgy but safe)
  while (readBytes >= 0)
    {
      readBytes += read (childWritePipe[READ_END], readIn + readBytes, 1);
      if (readBytes > 1024)
	{
	  fprintf (stderr, "Error: line is too long.\n");
	  return -1;
	}
      if (readBytes > 0 && readIn[readBytes - 1] == '\n')
	{
	  readIn[readBytes - 1] = '\0';
	  readBytes = -1;	// done
	}
    }

  // Check to see that there are 6 : delimeters in line
  readBytes = 0;		// reuse readBytes to count : in line
  for (i = strlen (readIn) - 1; i >= 0; i--)
    {
      if (readIn[i] == ':')
	{
	  readBytes++;
	  if (readBytes == 4)
	    readIn[i] = '\0';
	  else if (readBytes == 5)
	    memcpy (zoneState, readIn + i + 1, strlen (readIn + i + 1) + 1);
	}
    }

  if (readBytes == 6)
    return 0;
  return -1;

}
