#include "common.h"
#include "gpl.h"

void usage (int, char **);
int adm_init (int, char **);
int adm_add (int, char **);
int adm_del (int, char **);
int adm_create (int, char **);
int adm_free (int, char **);
int adm_list (int, char **);
int adm_uninit (int, char **);

int
main (int argc, char **argv)
{
  rootRequired ();
  init ();
  if (argc < 2 || strcmp (argv[1], "-h") == 0)
    {
      usage (argc, argv);
      return 0;
    }

#define RUN(a) if( strcmp( #a , argv[1]) == 0 ) { return adm_ ## a (argc, argv); }
  RUN (init);
  RUN (add);
  RUN (del);
  RUN (create);
  RUN (free);
  RUN (list);
  RUN (uninit);
#undef RUN
  // if nothing matches to run... show usage
  usage (argc, argv);
  return -1;
}

void
cleanup ()
{
  destroy ();
}

void
usage (int argc, char **argv)
{
  printf ("%s:\n", argv[0]);
  printf (" %s [-h]          - this message\n", argv[0]);
  printf (" %s init          - initialize the devzone system\n", argv[0]);
  printf (" %s add {zone}    - add a zone to the devzone list\n", argv[0]);
  printf (" %s del {zone}    - delete a zone from the devzone list\n",
	  argv[0]);
  printf (" %s create {zone} - create/reserve a devzone\n", argv[0]);
  printf (" %s free {zone}   - free a user reserved devzone\n", argv[0]);
  printf (" %s list [-p]     - list all devzones\n", argv[0]);
  printf (" %s uninit        - tear down the devzone system\n", argv[0]);
  cleanup ();
}

// Run a command and place stdout to a file
int
adm_run_file (char *status, char *command, char **argv, char *fn)
{
  int childPID, exitStatus;
  int print = 1;
  FILE *out;
  int out_fd;
  if (strlen (status) == 0)
    {
      print = 0;
    }
  if (print)
    printf ("Running %s...\n", status);
  if (fn != NULL)
    {
      out = fopen (fn, "w");
      if (!out)
	{
	  fprintf (stderr, "Error: could not open file for writing.\n");
	  cleanup ();
	  exit (-1);
	}
      out_fd = fileno (out);
    }
  childPID = fork ();
  // child PID
  if (!childPID)
    {
      if (fn != NULL)
	dup2 (out_fd, 1);
      dup2 (out_fd, 2);
      exit (execv (command, argv));
    }

  // parent lives on
  wait (&exitStatus);
  if (fn != NULL)
    {
      close (out_fd);
    }

  if (exitStatus != 0)
    {
      if (print)
	fprintf (stderr,
		 "Error: %s exited with status %d. Please resolve any errors.\n",
		 status, (exitStatus >> 8) & 255);
      return (exitStatus >> 8) & 255;
    }
  if (print)
    printf ("%s completed\n", status);
  return 0;
}

// Run a command
int
adm_run (char *status, char *command, char **argv)
{
  return adm_run_file (status, command, argv, NULL);
}

// Hand all initialization to devzone_init
int
adm_init (int argc, char **argv)
{
  char *args[2] = { "devzone_adm_init", (char *) NULL };

  // no need to keep DB open
  cleanup ();

  return adm_run ("devzone initialization",
		  "/etc/devzone/scripts/devzone_adm_init", args);
}


// Really long function that simply does:
/*

devzone_adm add {zone}
  -- check if zone already in arsenal
   - yes
     abort("{zone} already in devzone list")

  -- check if zonecfg exists
   - yes
     -- check if zone in state "configured"
      - yes
        zonecfg -z {zone} export > /etc/devzone/templates/{zone}
      - no
        abort("{zone} not in state configured")
   - no
     -- can "SCRIPT_BASE/devzone_adm_add_nc" figure out how to deal with this?
      - yes
        yay!
      - no
        abort("need a configured zone or parameters to configure a default zone")

*/
int
adm_add (int argc, char **argv)
{
  char *zone_name = argv[2];
  char *args[10];		// nothing seems to need more than 10 specified args
  int zonecfg_exists = 0;
  int zonecfg_state_configured = 0;
  struct stat zonecfg_stat;
  struct zone_info zonecfg_info;
  char *zone_cfg_path;
  char *zone_cfg_state_string;

  struct zone_info new_zone;

  // construct args
  args[0] = "zoneadm";
  args[1] = "-z";
  args[2] = zone_name;
  args[3] = "info";
  args[4] = "zonepath";
  args[5] = NULL;

  if (zone_name == NULL || strlen (zone_name) < 1
      || strlen (zone_name) >= CONFIG_ZONE_LEN)
    {
      fprintf (stderr, "Error: need to specify a valid zone name\n");
      usage (argc, argv);
      return -1;
    }

  // check if zone already in arsenal
  if (getZoneRecordByName (zone_name, &zonecfg_info) == 0)
    {
      fprintf (stderr, "Error: zone is already registered.\n");
      cleanup ();
      return -1;
    }

  // Do we need to supply a configuration file, or is one already there?
  // 25 == strlen( /etc/devzone/templates/ )
  zone_cfg_path = malloc (sizeof (char) * (CONFIG_ZONE_LEN + 25));
  sprintf (zone_cfg_path, "/etc/devzone/templates/%s", zone_name);
  if (stat (zone_cfg_path, &zonecfg_stat) != 0)
    {
      // Need a config file ... maybe zonecfg has something for us?

      zonecfg_exists =
	adm_run_file ("", "/usr/sbin/zonecfg", args, "/dev/null");
      if (zonecfg_exists == 0)
	{			// a configuration exists

	  // run grep -qi {zone}:configured /etc/zones/index to determine
	  // if the zone is in a configured state or not
	  // NB: this is a hack and can break at any time since /etc/zones/index is
	  //     not a published interface. zoneadm list -cp would be better but I'm
	  //     lazy and don't feel like debugging a bunch of pipes in C :) feel
	  //     free to improve my code.
	  zone_cfg_state_string =
	    (char *) malloc (sizeof (char) * (CONFIG_ZONE_LEN + 15));
	  sprintf (zone_cfg_state_string, "^%s:configured:", zone_name);

	  // construct args
	  args[0] = "grep";
	  args[1] = "-qi";
	  args[2] = zone_cfg_state_string;
	  args[3] = "/etc/zones/index";
	  args[4] = NULL;

	  zonecfg_state_configured =
	    adm_run ("zone state check", "/usr/bin/grep", args);

	  // zone state is not "configured" ... let's get out of here!
	  if (zonecfg_state_configured != 0)
	    {
	      fprintf (stderr, "Error: zone is not in state configured.\n");
	      cleanup ();
	      return -1;
	    }

	  // construct args
	  args[0] = "zonecfg";
	  args[1] = "-z";
	  args[2] = zone_name;
	  args[3] = "export";
	  args[4] = NULL;

	  // zonecfg -z {zone} export > /etc/devzone/config/{zone}
	  // re-use zonecfg_state_configured
	  zonecfg_state_configured =
	    adm_run_file ("", "/usr/sbin/zonecfg", args, zone_cfg_path);

	  if (zonecfg_state_configured != 0)
	    {
	      fprintf (stderr,
		       "Error: something went wrong while transferring zonecfg.\n");
	      return zonecfg_state_configured;
	    }

	}
      else
	{			// -- can "SCRIPT_BASE/devzone_adm_add_nc" figure out how to deal with this?

	  // if this fails it returns non-zero
	  // re-use zone_cfg_state_configured
	  zonecfg_state_configured =
	    adm_run ("", "/etc/devzone/scripts/devzone_adm_add_nc", argv);

	  if (zonecfg_state_configured != 0)
	    {
	      fprintf (stderr,
		       "Error: could not create a valid zone config. Please configure the zone.\n");
	      return zonecfg_state_configured;
	    }

	}

    }				// /etc/devzone/config/{zone} exists by now


  // -- add {zone} to devzone list
  strcpy (new_zone.name, zone_name);
  strcpy (new_zone.username, "");
  strcpy (new_zone.char_uid, "");
  new_zone.uid = -1;
  addRecord (&new_zone);

  cleanup ();
  return 0;
}

int
adm_del (int argc, char **argv)
{
  int ret;
  struct zone_info zi;

  if (argc < 2 || argv[2][0] == '\0' || strlen (argv[2]) > CONFIG_ZONE_LEN)
    {
      usage (argc, argv);
      cleanup ();
      return -1;
    }

  ret = getZoneRecordByName (argv[2], &zi);
  if (ret == -1)
    {
      //no such zone exists
      fprintf (stderr, "devzone %s does not exist\n", argv[2]);
      destroy ();
      return -1;
    }

  if (zi.uid != -1)
    {
      fprintf (stderr,
	       "devzone %s is currently reserved. First free it using %s free %s\n",
	       argv[2], argv[0], argv[2]);
      destroy ();
      return -1;
    }
  return delRecord (argv[2]);
}

int
adm_create (int argc, char **argv)
{
  int ret;
  struct zone_info zi;
  char *args[4] = { "init_zone.sh", "-zone", argv[2], NULL };

  // need a zone name of appropriate length (at least)
  if (argc < 2 || argv[2][0] == '\0' || strlen (argv[2]) > CONFIG_ZONE_LEN)
    {
      usage (argc, argv);
      cleanup ();
      return -1;
    }

  ret = getZoneRecordByName (argv[2], &zi);
  if (ret == -1)
    {
      //no such devzone exists
      fprintf (stderr, "devzone %s does not exist.\n", argv[2]);
      destroy ();
      return -1;
    }

  if (zi.uid != -1)
    {
      fprintf (stderr,
	       "devzone %s (%d) is already reserved. First free it using %s free %s\n",
	       argv[2], zi.uid, argv[0], argv[2]);
      destroy ();
      return -1;
    }

  reserveZone (argv[2]);

  ret = adm_run ("", "/etc/devzone/scripts/init_zone.sh", args);

  return ret;
}

int
adm_free (int argc, char **argv)
{
  char *args[4] = { "release_zone.sh", "-zone", argv[2], (char *) NULL };
  int ret;

  ret =
    adm_run ("Releasing zone", "/etc/devzone/scripts/release_zone.sh", args);

  if (ret != 0)
    {
      cleanup ();
      fprintf (stderr, "Error: could not free zone.\n");
      return -1;
    }
  freeZone (argv[2]);
  cleanup ();
  return 0;
}

void
adm_list_human (struct zone_info **zone_record)
{
  int zone_length = 4;		// strlen('zone') == 4
  int user_length = 4;		// strlen('user') == 4
  int i = 0;

  int c_zl, c_ul;

  while (zone_record[i] != NULL)
    {
      zone_length =
	(strlen (zone_record[i]->name) >
	 zone_length) ? strlen (zone_record[i]->name) : zone_length;
      user_length =
	(strlen (zone_record[i]->username) >
	 user_length) ? strlen (zone_record[i]->username) : user_length;
      i++;
    }

#define COLUMN_PAD 2

  // Print line header
  printf ("Zone");
  for (c_zl = zone_length - 4 + COLUMN_PAD; c_zl > 0; c_zl--)
    printf (" ");

  printf ("User");
  for (c_ul = user_length - 4 + COLUMN_PAD; c_ul > 0; c_ul--)
    printf (" ");

  printf ("UID\n");

  i = 0;

  // print records
  while (zone_record[i] != NULL)
    {
      c_zl = strlen (zone_record[i]->name);
      c_ul = strlen (zone_record[i]->username);

      printf ("%s", zone_record[i]->name);

      // zone padding
      for (c_zl = zone_length - c_zl + COLUMN_PAD; c_zl > 0; c_zl--)
	printf (" ");

      printf ("%s", zone_record[i]->username);

      // user padding
      for (c_ul = user_length - c_ul + COLUMN_PAD; c_ul > 0; c_ul--)
	printf (" ");

      printf ("%d\n", zone_record[i]->uid);

      i++;
    }

}

int
adm_list (int argc, char **argv)
{
  struct zone_info **zone_list;
  int count = 0;

  // NULL terminated list
  zone_list = getRecordList ();

  // parsable output?
  if (argc > 2 && strcmp (argv[2], "-p") == 0)
    {

      while (zone_list[count] != NULL)
	{
	  printf ("%s:%s:%d\n", zone_list[count]->name,
		  zone_list[count]->username, zone_list[count]->uid);
	  count++;
	}
      free (zone_list);
      cleanup ();
      return 0;
    }

  // Do we generate human readable output?
  adm_list_human (zone_list);

  free (zone_list);
  cleanup ();
  return 0;
}

int
adm_uninit (int argc, char **argv)
{
  printf ("Running uninit...\n");
  cleanup ();
  return 0;
}
